"""
Builder for AgentConfig objects using a fluent interface.
"""

import copy

from langchain_core.callbacks.base import Callbacks

from minitap.mobile_use.config import get_default_llm_config, get_default_minitap_llm_config
from minitap.mobile_use.context import DevicePlatform
from minitap.mobile_use.sdk.constants import DEFAULT_PROFILE_NAME
from minitap.mobile_use.sdk.types.agent import AgentConfig, AgentProfile, ServerConfig
from minitap.mobile_use.sdk.types.task import TaskRequestCommon


class AgentConfigBuilder:
    """
    Builder class providing a fluent interface for creating AgentConfig objects.

    This builder allows for step-by-step construction of an AgentConfig with
    clear methods that make the configuration process intuitive and type-safe.

    Examples:
        >>> builder = AgentConfigBuilder()
        >>> config = (builder
        ...     .add_profile(AgentProfile(name="HighReasoning", llm_config=LLMConfig(...)))
        ...     .add_profile(AgentProfile(name="LowReasoning", llm_config=LLMConfig(...)))
        ...     .for_device(DevicePlatform.ANDROID, "device123")
        ...     .with_default_task_config(TaskRequestCommon(max_steps=30))
        ...     .with_default_profile("HighReasoning")
        ...     .build()
        ... )
    """

    def __init__(self):
        """Initialize an empty AgentConfigBuilder."""
        self._agent_profiles: dict[str, AgentProfile] = {}
        self._task_request_defaults: TaskRequestCommon | None = None
        self._default_profile: str | AgentProfile | None = None
        self._device_id: str | None = None
        self._device_platform: DevicePlatform | None = None
        self._servers: ServerConfig = get_default_servers()
        self._graph_config_callbacks: Callbacks = None
        self._cloud_mobile_id_or_ref: str | None = None

    def add_profile(self, profile: AgentProfile, validate: bool = True) -> "AgentConfigBuilder":
        """
        Add an agent profile to the mobile-use agent.

        Args:
            profile: The agent profile to add
        """
        self._agent_profiles[profile.name] = profile
        if validate:
            profile.llm_config.validate_providers()
        return self

    def add_profiles(
        self,
        profiles: list[AgentProfile],
        validate: bool = True,
    ) -> "AgentConfigBuilder":
        """
        Add multiple agent profiles to the mobile-use agent.

        Args:
            profiles: List of agent profiles to add
        """
        for profile in profiles:
            self.add_profile(profile=profile, validate=validate)
        return self

    def with_default_profile(self, profile: str | AgentProfile) -> "AgentConfigBuilder":
        """
        Set the default agent profile used for tasks.

        Args:
            profile: The name or instance of the default agent profile
        """
        self._default_profile = profile
        return self

    def for_device(
        self,
        platform: DevicePlatform,
        device_id: str,
    ) -> "AgentConfigBuilder":
        """
        Configure the mobile-use agent for a specific device.

        Args:
            platform: The device platform (ANDROID or IOS)
            device_id: The unique identifier for the device
        """
        if self._cloud_mobile_id_or_ref is not None:
            raise ValueError(
                "Device ID cannot be set when a cloud mobile is already configured.\n"
                "> for_device() and for_cloud_mobile() are mutually exclusive"
            )
        self._device_id = device_id
        self._device_platform = platform
        return self

    def for_cloud_mobile(self, cloud_mobile_id_or_ref: str) -> "AgentConfigBuilder":
        """
        Configure the mobile-use agent to use a cloud mobile.

        When using a cloud mobile, tasks are executed remotely via the Platform API,
        and only PlatformTaskRequest can be used.

        Args:
            cloud_mobile_id_or_ref: The unique identifier or reference name for the cloud mobile.
                Can be either a UUID (e.g., '550e8400-e29b-41d4-a716-446655440000')
                or a reference name (e.g., 'my-test-device')
        """
        if self._device_id is not None:
            raise ValueError(
                "Cloud mobile device ID cannot be set when a device is already configured.\n"
                "> for_device() and for_cloud_mobile() are mutually exclusive"
            )
        self._cloud_mobile_id_or_ref = cloud_mobile_id_or_ref
        return self

    def with_default_task_config(self, config: TaskRequestCommon) -> "AgentConfigBuilder":
        """
        Set the default task configuration.

        Args:
            config: The task configuration to use as default
        """
        self._task_request_defaults = copy.deepcopy(config)
        return self

    def with_adb_server(self, host: str, port: int | None = None) -> "AgentConfigBuilder":
        """
        Set the ADB server host and port.

        Args:
            host: The ADB server host
            port: The ADB server port
        """
        self._servers.adb_host = host
        if port is not None:
            self._servers.adb_port = port
        return self

    def with_servers(self, servers: ServerConfig) -> "AgentConfigBuilder":
        """
        Set the server settings.

        Args:
            servers: The server settings to use
        """
        self._servers = copy.deepcopy(servers)
        return self

    def with_graph_config_callbacks(self, callbacks: Callbacks) -> "AgentConfigBuilder":
        """
        Set the graph config callbacks.

        Args:
            callbacks: The graph config callbacks to use
        """
        self._graph_config_callbacks = callbacks
        return self

    def build(self, validate_profiles: bool = True) -> AgentConfig:
        """
        Build the mobile-use AgentConfig object.

        Args:
            default_profile: Name of the default agent profile to use

        Returns:
            A configured AgentConfig object

        Raises:
            ValueError: If default_profile is specified but not found in configured profiles
        """
        nb_profiles = len(self._agent_profiles)

        if isinstance(self._default_profile, str):
            profile_name = self._default_profile
            default_profile = self._agent_profiles.get(profile_name, None)
            if default_profile is None:
                raise ValueError(f"Profile '{profile_name}' not found in configured agents")
        elif isinstance(self._default_profile, AgentProfile):
            default_profile = self._default_profile
            if default_profile.name not in self._agent_profiles:
                self.add_profile(default_profile, validate=validate_profiles)
        elif nb_profiles <= 0:
            llm_config = (
                get_default_minitap_llm_config(validate=validate_profiles)
                or get_default_llm_config()
            )
            default_profile = AgentProfile(
                name=DEFAULT_PROFILE_NAME,
                llm_config=llm_config,
            )
            self.add_profile(default_profile, validate=validate_profiles)
        elif nb_profiles == 1:
            # Select the only one available
            default_profile = next(iter(self._agent_profiles.values()))
        else:
            available_profiles = ", ".join(self._agent_profiles.keys())
            raise ValueError(
                f"You must call with_default_profile() to select one among: {available_profiles}"
            )

        return AgentConfig(
            agent_profiles=self._agent_profiles,
            task_request_defaults=self._task_request_defaults or TaskRequestCommon(),
            default_profile=default_profile,
            device_id=self._device_id,
            device_platform=self._device_platform,
            servers=self._servers,
            graph_config_callbacks=self._graph_config_callbacks,
            cloud_mobile_id_or_ref=self._cloud_mobile_id_or_ref,
        )


def get_default_agent_config():
    return AgentConfigBuilder().build()


def get_default_servers():
    return ServerConfig(
        adb_host="localhost",
        adb_port=5037,
    )



import asyncio
import sys
import tempfile
import uuid
from collections.abc import Callable, Coroutine
from datetime import UTC, datetime
from io import BytesIO
from pathlib import Path
from shutil import which
from types import NoneType
from typing import Any, TypeVar, overload

from adbutils import AdbClient
from dotenv import load_dotenv
from langchain_core.messages import AIMessage
from PIL import Image
from pydantic import BaseModel

from minitap.mobile_use.agents.outputter.outputter import outputter
from minitap.mobile_use.agents.planner.types import Subgoal
from minitap.mobile_use.clients.idb_client import IdbClientWrapper
from minitap.mobile_use.clients.ui_automator_client import UIAutomatorClient
from minitap.mobile_use.config import AgentNode, OutputConfig, record_events, settings
from minitap.mobile_use.context import (
    DeviceContext,
    DevicePlatform,
    ExecutionSetup,
    IsReplan,
    MobileUseContext,
)
from minitap.mobile_use.controllers.platform_specific_commands_controller import get_first_device
from minitap.mobile_use.graph.graph import get_graph
from minitap.mobile_use.graph.state import State
from minitap.mobile_use.sdk.builders.agent_config_builder import get_default_agent_config
from minitap.mobile_use.sdk.builders.task_request_builder import TaskRequestBuilder
from minitap.mobile_use.sdk.services.cloud_mobile import CloudMobileService
from minitap.mobile_use.sdk.services.platform import PlatformService
from minitap.mobile_use.sdk.types.agent import AgentConfig
from minitap.mobile_use.sdk.types.exceptions import (
    AgentNotInitializedError,
    AgentProfileNotFoundError,
    AgentTaskRequestError,
    CloudMobileServiceUninitializedError,
    DeviceNotFoundError,
    ExecutableNotFoundError,
    PlatformServiceUninitializedError,
    ServerStartupError,
)
from minitap.mobile_use.sdk.types.platform import TaskRunPlanResponse, TaskRunStatus
from minitap.mobile_use.sdk.types.task import (
    AgentProfile,
    CloudDevicePlatformTaskRequest,
    PlatformTaskInfo,
    PlatformTaskRequest,
    Task,
    TaskRequest,
)
from minitap.mobile_use.utils.app_launch_utils import _handle_initial_app_launch
from minitap.mobile_use.utils.logger import get_logger
from minitap.mobile_use.utils.media import (
    create_gif_from_trace_folder,
    create_steps_json_from_trace_folder,
    remove_images_from_trace_folder,
    remove_steps_json_from_trace_folder,
)
from minitap.mobile_use.utils.recorder import log_agent_thought

logger = get_logger(__name__)

TOutput = TypeVar("TOutput", bound=BaseModel | None)

load_dotenv()


class Agent:
    _config: AgentConfig
    _tasks: list[Task] = []
    _tmp_traces_dir: Path
    _initialized: bool = False
    _device_context: DeviceContext
    _adb_client: AdbClient | None
    _ui_adb_client: UIAutomatorClient | None
    _idb_client: IdbClientWrapper | None
    _current_task: asyncio.Task | None = None
    _task_lock: asyncio.Lock
    _cloud_mobile_id: str | None = None

    def __init__(self, *, config: AgentConfig | None = None):
        self._config = config or get_default_agent_config()
        self._tasks = []
        self._tmp_traces_dir = Path(tempfile.gettempdir()) / "mobile-use-traces"
        self._initialized = False
        self._task_lock = asyncio.Lock()

        # Initialize platform service if API key is available in environment
        # Note: Can also be initialized later with API key at agent .init()
        if settings.MINITAP_API_KEY:
            self._platform_service = PlatformService()
            self._cloud_mobile_service = CloudMobileService()
        else:
            self._platform_service = None
            self._cloud_mobile_service = None

    async def init(
        self,
        api_key: str | None = None,
        server_restart_attempts: int = 3,
        retry_count: int = 5,
        retry_wait_seconds: int = 5,
    ):
        if api_key:
            self._platform_service = PlatformService(api_key=api_key)
            self._cloud_mobile_service = CloudMobileService(api_key=api_key)

        # Skip initialization for cloud devices - no local setup required
        if self._config.cloud_mobile_id_or_ref:
            if not self._cloud_mobile_service:
                raise CloudMobileServiceUninitializedError()
            self._cloud_mobile_id = await self._cloud_mobile_service.resolve_cloud_mobile_id(
                cloud_mobile_id_or_ref=self._config.cloud_mobile_id_or_ref,
            )
            logger.info("Cloud device configured - skipping local initialization")
            self._initialized = True
            return True

        if not which("adb") and not which("xcrun"):
            raise ExecutableNotFoundError("cli_tools")

        if self._initialized:
            logger.warning("Agent is already initialized. Skipping...")
            return True

        # Get first available device ID
        if not self._config.device_id or not self._config.device_platform:
            device_id, platform = get_first_device(logger=logger)
        else:
            device_id, platform = self._config.device_id, self._config.device_platform

        if not device_id or not platform:
            error_msg = "No device found. Exiting."
            logger.error(error_msg)
            raise DeviceNotFoundError(error_msg)

        # Initialize clients
        self._init_clients(
            device_id=device_id,
            platform=platform,
            retry_count=retry_count,
            retry_wait_seconds=retry_wait_seconds,
        )

        # Initialize IDB companion for iOS
        if self._idb_client:
            logger.info("Starting IDB companion for iOS device...")
            companion_started = await self._idb_client.init_companion()
            if not companion_started:
                raise ServerStartupError(
                    message="Failed to start IDB companion for iOS device. "
                    "Please ensure fb-idb is installed: https://fbidb.io/docs/installation/"
                )
            logger.success("IDB companion started successfully")

        # Start necessary servers
        restart_attempt = 0
        while restart_attempt < server_restart_attempts:
            success = self._run_servers(
                device_id=device_id,
                platform=platform,
            )
            if success:
                break

            restart_attempt += 1
            if restart_attempt < server_restart_attempts:
                logger.warning(
                    f"Server start failed, attempting restart "
                    f"{restart_attempt}/{server_restart_attempts}"
                )
            else:
                error_msg = "Mobile-use servers failed to start after all restart attempts."
                logger.error(error_msg)
                raise ServerStartupError(message=error_msg)

        self._device_context = await self._get_device_context(
            device_id=device_id, platform=platform
        )
        logger.info(self._device_context.to_str())
        logger.info("✅ Mobile-use agent initialized.")
        self._initialized = True
        return True

    def new_task(self, goal: str):
        """
        Create a new task request builder.

        Args:
            goal: Natural language description of what to accomplish

        Returns:
            TaskRequestBuilder that can be configured with:
            - .with_output_format() for structured output
            - .with_output_description() for output description
            - .with_locked_app_package() to restrict execution to a specific app
            - .using_profile() to specify an LLM profile
            - .with_max_steps() to set maximum execution steps
            - .with_trace_recording() to enable trace recording
            - .with_name() to set a custom task name
        """
        return TaskRequestBuilder[None].from_common(
            goal=goal,
            common=self._config.task_request_defaults,
        )

    @overload
    async def run_task(
        self,
        *,
        goal: str,
        output: type[TOutput],
        profile: str | AgentProfile | None = None,
        name: str | None = None,
        locked_app_package: str | None = None,
    ) -> TOutput | None: ...

    @overload
    async def run_task(
        self,
        *,
        goal: str,
        output: str,
        profile: str | AgentProfile | None = None,
        name: str | None = None,
        locked_app_package: str | None = None,
    ) -> str | dict | None: ...

    @overload
    async def run_task(
        self,
        *,
        goal: str,
        output=None,
        profile: str | AgentProfile | None = None,
        name: str | None = None,
        locked_app_package: str | None = None,
    ) -> str | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: TaskRequest[None],
        locked_app_package: str | None = None,
    ) -> str | dict | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: TaskRequest[TOutput],
        locked_app_package: str | None = None,
    ) -> TOutput | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: PlatformTaskRequest[None],
        locked_app_package: str | None = None,
    ) -> str | dict | None: ...

    @overload
    async def run_task(
        self,
        *,
        request: PlatformTaskRequest[TOutput],
        locked_app_package: str | None = None,
    ) -> TOutput | None: ...

    async def run_task(
        self,
        *,
        goal: str | None = None,
        output: type[TOutput] | str | None = None,
        profile: str | AgentProfile | None = None,
        locked_app_package: str | None = None,
        name: str | None = None,
        request: TaskRequest[TOutput] | PlatformTaskRequest[TOutput] | None = None,
    ) -> str | dict | TOutput | None:
        # Check if cloud mobile is configured
        if self._config.cloud_mobile_id_or_ref:
            if request is None or not isinstance(request, PlatformTaskRequest):
                raise AgentTaskRequestError(
                    "When using a cloud mobile, only PlatformTaskRequest is supported. "
                    "Use AgentConfigBuilder.for_cloud_mobile() only with PlatformTaskRequest."
                )
            # Use cloud mobile execution path
            return await self._run_cloud_mobile_task(
                request=request, locked_app_package=locked_app_package
            )

        # Normal local execution path
        if request is not None:
            task_info = None
            if isinstance(request, PlatformTaskRequest):
                if not self._platform_service:
                    raise PlatformServiceUninitializedError()
                task_info = await self._platform_service.create_task_run(
                    request=request, locked_app_package=locked_app_package
                )
                if isinstance(request, CloudDevicePlatformTaskRequest):
                    request.task_run_id = task_info.task_run.id
                    request.task_run_id_available_event.set()
                self._config.agent_profiles[task_info.llm_profile.name] = task_info.llm_profile
                request = task_info.task_request
            elif locked_app_package is not None:
                if request.locked_app_package:
                    logger.warning(
                        "Locked app package specified both in the request and as a parameter. "
                        "Using the parameter value."
                    )
                request.locked_app_package = locked_app_package
            return await self._run_task(
                request=request, task_info=task_info, platform_service=self._platform_service
            )
        if goal is None:
            raise AgentTaskRequestError("Goal is required")
        task_request = self.new_task(goal=goal)
        if output is not None:
            if isinstance(output, str):
                task_request.with_output_description(description=output)
            elif output is not NoneType:
                task_request.with_output_format(output_format=output)
        if profile is not None:
            task_request.using_profile(profile=profile)
        if name is not None:
            task_request.with_name(name=name)
        if locked_app_package is not None:
            task_request.with_locked_app_package(package_name=locked_app_package)
        return await self._run_task(task_request.build())

    async def _run_cloud_mobile_task(
        self,
        request: PlatformTaskRequest[TOutput],
        locked_app_package: str | None = None,
    ) -> str | dict | TOutput | None:
        """
        Execute a task on a cloud mobile.

        This method triggers the task execution on the Platform and polls
        for completion without running any agentic logic locally.
        """
        if not self._cloud_mobile_id:
            raise AgentTaskRequestError("Cloud mobile ID is not configured")

        if not self._cloud_mobile_service:
            raise CloudMobileServiceUninitializedError()

        # Start cloud mobile if not already started
        logger.info(f"Starting cloud mobile '{self._cloud_mobile_id}'...")
        await self._cloud_mobile_service.start_and_wait_for_ready(
            cloud_mobile_id=self._cloud_mobile_id,
        )
        logger.info(
            f"Starting cloud mobile task execution '{self._cloud_mobile_id}'",
        )

        def log_callback(message: str):
            """Callback for logging timeline updates."""
            logger.info(message)

        def status_callback(
            status: TaskRunStatus,
            status_message: str | None,
        ):
            """Callback for status updates."""
            logger.info(f"Task status update: [{status}] {status_message}")

        async def _execute_cloud(cloud_mobile_service: CloudMobileService, cloud_mobile_id: str):
            try:
                # Execute task on cloud mobile and wait for completion
                final_status, error, output = await cloud_mobile_service.run_task_on_cloud_mobile(
                    cloud_mobile_id=cloud_mobile_id,
                    request=request,
                    on_status_update=status_callback,
                    on_log=log_callback,
                    locked_app_package=locked_app_package,
                )
                if final_status == "completed":
                    logger.success("Cloud mobile task completed successfully")
                    return output
                if final_status == "failed":
                    logger.error(f"Cloud mobile task failed: {error}")
                    raise AgentTaskRequestError(
                        f"Task execution failed on cloud mobile: {error}",
                    )
                if final_status == "cancelled":
                    logger.warning("Cloud mobile task was cancelled")
                    raise AgentTaskRequestError("Task execution was cancelled")
                logger.error(f"Unknown cloud mobile task status: {final_status}")
                raise AgentTaskRequestError(f"Unknown task status: {final_status}")
            except asyncio.CancelledError:
                # Propagate cancellation to parent coroutine.
                logger.info("Task cancelled during execution, re-raising CancelledError")
                raise
            except AgentTaskRequestError:
                # Re-raise known exceptions
                raise
            except Exception as e:
                logger.error(f"Unexpected error during cloud mobile task execution: {e}")
                raise AgentTaskRequestError(f"Unexpected error: {e}") from e

        async with self._task_lock:
            if self._current_task and not self._current_task.done():
                logger.warning(
                    "Another cloud task is running; cancelling it before starting new one.",
                )
                self.stop_current_task()
                try:
                    await self._current_task
                except asyncio.CancelledError:
                    pass
            try:
                self._current_task = asyncio.create_task(
                    _execute_cloud(
                        cloud_mobile_service=self._cloud_mobile_service,
                        cloud_mobile_id=self._cloud_mobile_id,
                    ),
                )
                return await self._current_task
            finally:
                self._current_task = None

    async def _run_task(
        self,
        request: TaskRequest[TOutput],
        task_info: PlatformTaskInfo | None = None,
        platform_service: PlatformService | None = None,
    ) -> str | dict | TOutput | None:
        if not self._initialized:
            raise AgentNotInitializedError()

        if request.profile:
            agent_profile = self._config.agent_profiles.get(request.profile)
            if agent_profile is None:
                raise AgentProfileNotFoundError(request.profile)
        else:
            agent_profile = self._config.default_profile
        logger.info(str(agent_profile))

        on_status_changed = None
        on_agent_thought = None
        on_plan_changes = None
        task_id = str(uuid.uuid4())
        if task_info:
            on_status_changed = self._get_task_status_change_callback(
                task_info=task_info, platform_service=platform_service
            )
            on_agent_thought = self._get_new_agent_thought_callback(
                task_info=task_info, platform_service=platform_service
            )
            on_plan_changes = self._get_plan_changes_callback(
                task_info=task_info, platform_service=platform_service
            )
            task_id = task_info.task_run.id

        task = Task(
            id=task_id,
            device=self._device_context,
            status="pending",
            request=request,
            created_at=datetime.now(),
            on_status_changed=on_status_changed,
        )
        self._tasks.append(task)
        task_name = task.get_name()

        # Extract API key from platform service if available
        api_key = None
        if platform_service:
            api_key = platform_service._api_key

        context = MobileUseContext(
            trace_id=task.id,
            device=self._device_context,
            adb_client=self._adb_client,
            ui_adb_client=self._ui_adb_client,
            idb_client=self._idb_client,
            llm_config=agent_profile.llm_config,
            on_agent_thought=on_agent_thought,
            on_plan_changes=on_plan_changes,
            minitap_api_key=api_key,
        )

        self._prepare_tracing(task=task, context=context)
        await self._prepare_app_lock(task=task, context=context)
        self._prepare_output_files(task=task)

        output_config = None
        if request.output_description or request.output_format:
            output_config = OutputConfig(
                output_description=request.output_description,
                structured_output=request.output_format,  # type: ignore
            )
            logger.info(str(output_config))

        logger.info(f"[{task_name}] Starting graph with goal: `{request.goal}`")
        state = self._get_graph_state(task=task)
        graph_input = state.model_dump()

        async def _execute_task_logic():
            last_state: State | None = None
            last_state_snapshot: dict | None = None
            output = None
            try:
                logger.info(f"[{task_name}] Invoking graph with input: {graph_input}")
                await task.set_status(status="running", message="Invoking graph...")
                async for chunk in (await get_graph(context)).astream(
                    input=graph_input,
                    config={
                        "recursion_limit": task.request.max_steps,
                        "callbacks": self._config.graph_config_callbacks,
                    },
                    stream_mode=["messages", "custom", "updates", "values"],
                ):
                    stream_mode, payload = chunk
                    if stream_mode == "values":
                        last_state_snapshot = payload  # type: ignore
                        last_state = State(**last_state_snapshot)  # type: ignore
                        if task.request.thoughts_output_path:
                            record_events(
                                output_path=task.request.thoughts_output_path,
                                events=last_state.agents_thoughts,
                            )

                    if stream_mode == "updates":
                        for _, value in payload.items():  # type: ignore node name, node output
                            if value and "agents_thoughts" in value:
                                new_thoughts = value["agents_thoughts"]
                                last_item = new_thoughts[-1] if new_thoughts else None
                                if last_item:
                                    log_agent_thought(
                                        agent_thought=last_item,
                                    )

                if not last_state:
                    err = f"[{task_name}] No result received from graph"
                    logger.warning(err)
                    await task.finalize(content=output, state=last_state_snapshot, error=err)
                    return None

                print_ai_response_to_stderr(graph_result=last_state)
                output = await self._extract_output(
                    task_name=task_name,
                    ctx=context,
                    request=request,
                    output_config=output_config,
                    state=last_state,
                )
                logger.info(f"✅ Automation '{task_name}' is success ✅")
                await task.finalize(content=output, state=last_state_snapshot)
                return output
            except asyncio.CancelledError:
                err = f"[{task_name}] Task cancelled"
                logger.warning(err)
                await task.finalize(
                    content=output,
                    state=last_state_snapshot,
                    error=err,
                    cancelled=True,
                )
                raise
            except Exception as e:
                err = f"[{task_name}] Error running automation: {e}"
                logger.error(err)
                await task.finalize(
                    content=output,
                    state=last_state_snapshot,
                    error=err,
                )
                raise
            finally:
                await self._finalize_tracing(task=task, context=context)

        async with self._task_lock:
            if self._current_task and not self._current_task.done():
                logger.warning(
                    "Another automation task is already running. "
                    "Stopping it before starting the new one."
                )
                self.stop_current_task()
                try:
                    await self._current_task
                except asyncio.CancelledError:
                    pass

            try:
                self._current_task = asyncio.create_task(_execute_task_logic())
                return await self._current_task
            finally:
                self._current_task = None

    def stop_current_task(self):
        """Requests cancellation of the currently running automation task."""
        if self._current_task and not self._current_task.done():
            logger.info("Requesting to stop the current automation task...")
            was_cancelled = self._current_task.cancel()
            if was_cancelled:
                logger.success("Cancellation request for the current task was sent.")
            else:
                logger.warning(
                    "Could not send cancellation request for the current task "
                    "(it may already be completing)."
                )
        else:
            logger.info("No active automation task to stop.")

    async def get_screenshot(self) -> Image.Image:
        """
        Capture a screenshot from the mobile device.

        For cloud mobiles, this method calls the mobile-manager endpoint.
        For local mobiles, it uses ADB (Android) or xcrun (iOS) directly.

        Returns:
            Screenshot as PIL Image

        Raises:
            AgentNotInitializedError: If the agent is not initialized
            PlatformServiceUninitializedError: If cloud mobile service is not available
            Exception: If screenshot capture fails
        """
        # Check if cloud mobile is configured
        if self._cloud_mobile_id:
            if not self._cloud_mobile_service:
                raise CloudMobileServiceUninitializedError()
            screenshot = await self._cloud_mobile_service.get_screenshot(
                cloud_mobile_id=self._cloud_mobile_id,
            )
            return screenshot

        # Local device - use ADB or xcrun directly
        if not self._initialized:
            raise AgentNotInitializedError()

        if self._device_context.mobile_platform == DevicePlatform.ANDROID:
            # Use ADB to capture screenshot
            logger.info("Capturing screenshot from local Android device")
            if not self._adb_client:
                raise Exception("ADB client not initialized")

            device = self._adb_client.device(serial=self._device_context.device_id)
            screenshot = await asyncio.to_thread(device.screenshot)
            logger.info("Screenshot captured from local Android device")
            return screenshot

        elif self._device_context.mobile_platform == DevicePlatform.IOS:
            # Use xcrun to capture screenshot
            import functools
            import subprocess
            from io import BytesIO

            logger.info("Capturing screenshot from local iOS device")
            try:
                # xcrun simctl io <device> screenshot --type=png -
                result = await asyncio.to_thread(
                    functools.partial(
                        subprocess.run,
                        [
                            "xcrun",
                            "simctl",
                            "io",
                            self._device_context.device_id,
                            "screenshot",
                            "--type=png",
                            "-",
                        ],
                        capture_output=True,
                        check=True,
                    )
                )
                # Convert bytes to PIL Image
                screenshot = Image.open(BytesIO(result.stdout))
                logger.info("Screenshot captured from local iOS device")
                return screenshot
            except subprocess.CalledProcessError as e:
                logger.error(f"Failed to capture screenshot: {e}")
                raise Exception(f"Failed to capture screenshot from iOS device: {e}")

        else:
            raise Exception(f"Unsupported platform: {self._device_context.mobile_platform}")

    async def clean(self, force: bool = False):
        if self._cloud_mobile_id:
            self._initialized = False
            logger.info("✅ Cloud-mode agent stopped.")
            return
        if not self._initialized and not force:
            return

        if self._idb_client:
            await self._idb_client.cleanup()
            self._idb_client = None

        self._initialized = False
        logger.info("✅ Mobile-use agent stopped.")

    async def _prepare_app_lock(self, task: Task, context: MobileUseContext):
        """Prepare app lock by launching the locked app if specified."""
        if not task.request.locked_app_package:
            return

        task_name = task.get_name()
        logger.info(f"[{task_name}] Preparing app lock for: {task.request.locked_app_package}")

        app_lock_status = await _handle_initial_app_launch(
            ctx=context, locked_app_package=task.request.locked_app_package
        )

        if context.execution_setup is None:
            context.execution_setup = ExecutionSetup(app_lock_status=app_lock_status)
        else:
            context.execution_setup.app_lock_status = app_lock_status

        if app_lock_status.locked_app_initial_launch_success is False:
            error = app_lock_status.locked_app_initial_launch_error
            logger.warning(f"[{task_name}] Failed to launch locked app: {error}")

    def _prepare_tracing(self, task: Task, context: MobileUseContext):
        """Prepare tracing setup if record_trace is enabled."""
        if not task.request.record_trace:
            return

        task_name = task.get_name()
        temp_trace_path = Path(self._tmp_traces_dir / task_name).resolve()
        traces_output_path = Path(task.request.trace_path).resolve()
        logger.info(f"[{task_name}] 📂 Traces output path: {traces_output_path}")
        logger.info(f"[{task_name}] 📄📂 Traces temp path: {temp_trace_path}")
        traces_output_path.mkdir(parents=True, exist_ok=True)
        temp_trace_path.mkdir(parents=True, exist_ok=True)

        context.execution_setup = ExecutionSetup(
            traces_path=self._tmp_traces_dir,
            trace_name=task_name,
            enable_remote_tracing=task.request.enable_remote_tracing,
        )

    async def _finalize_tracing(self, task: Task, context: MobileUseContext):
        exec_setup_ctx = context.execution_setup
        if not exec_setup_ctx:
            return

        if exec_setup_ctx.traces_path is None or exec_setup_ctx.trace_name is None:
            return

        task_name = task.get_name()
        status = "_PASS" if task.status == "completed" else "_FAIL"
        ts = task.created_at.strftime("%Y-%m-%dT%H-%M-%S")
        new_name = f"{exec_setup_ctx.trace_name}{status}_{ts}"

        temp_trace_path = (self._tmp_traces_dir / exec_setup_ctx.trace_name).resolve()
        traces_output_path = Path(task.request.trace_path).resolve()

        logger.info(f"[{task_name}] Compiling trace FROM FOLDER: " + str(temp_trace_path))
        create_gif_from_trace_folder(temp_trace_path)
        create_steps_json_from_trace_folder(temp_trace_path)

        if exec_setup_ctx.enable_remote_tracing:
            gif_path = temp_trace_path / "trace.gif"
            if gif_path.exists() and self._platform_service:
                try:
                    task_run_id = await self._platform_service.upload_trace_gif(
                        task_run_id=task.id, gif_path=gif_path
                    )
                    if task_run_id:
                        platform_url = f"{settings.MINITAP_BASE_URL}/task-runs/{task_run_id}"
                        logger.info(f"[{task_name}] 🌐 View on platform: {platform_url}")
                except Exception as e:
                    logger.warning(f"[{task_name}] Failed to upload trace GIF: {e}")

        logger.info(f"[{task_name}] Video created, removing dust...")
        remove_images_from_trace_folder(temp_trace_path)
        remove_steps_json_from_trace_folder(temp_trace_path)
        logger.info(f"[{task_name}] 📽️ Trace compiled, moving to output path 📽️")

        output_folder_path = temp_trace_path.rename(traces_output_path / new_name).resolve()
        logger.info(f"[{task_name}] 📂✅ Traces located in: {output_folder_path}")

    def _prepare_output_files(self, task: Task):
        if task.request.llm_output_path:
            _validate_and_prepare_file(file_path=task.request.llm_output_path)
        if task.request.thoughts_output_path:
            _validate_and_prepare_file(file_path=task.request.thoughts_output_path)

    async def _extract_output(
        self,
        task_name: str,
        ctx: MobileUseContext,
        request: TaskRequest[TOutput],
        output_config: OutputConfig | None,
        state: State,
    ) -> str | dict | TOutput | None:
        if output_config and output_config.needs_structured_format():
            logger.info(f"[{task_name}] Generating structured output...")
            try:
                structured_output = await outputter(
                    ctx=ctx,
                    output_config=output_config,
                    graph_output=state,
                )
                logger.info(f"[{task_name}] Structured output: {structured_output}")
                record_events(output_path=request.llm_output_path, events=structured_output)
                if request.output_format is not None and request.output_format is not NoneType:
                    return request.output_format.model_validate(structured_output)
                return structured_output
            except Exception as e:
                logger.error(f"[{task_name}] Failed to generate structured output: {e}")
                return None
        if state and state.agents_thoughts:
            last_msg = state.agents_thoughts[-1]
            logger.info(str(last_msg))
            record_events(output_path=request.llm_output_path, events=last_msg)
            return last_msg
        return None

    def _get_graph_state(self, task: Task):
        return State(
            messages=[],
            initial_goal=task.request.goal,
            subgoal_plan=[],
            latest_ui_hierarchy=None,
            latest_screenshot=None,
            focused_app_info=None,
            device_date=None,
            structured_decisions=None,
            complete_subgoals_by_ids=[],
            agents_thoughts=[],
            remaining_steps=task.request.max_steps,
            executor_messages=[],
            cortex_last_thought=None,
        )

    def _init_clients(
        self,
        device_id: str,
        platform: DevicePlatform,
        retry_count: int,
        retry_wait_seconds: int,
    ):
        self._adb_client = (
            AdbClient(host=self._config.servers.adb_host, port=self._config.servers.adb_port)
            if platform == DevicePlatform.ANDROID
            else None
        )
        self._ui_adb_client = (
            UIAutomatorClient(device_id=device_id) if platform == DevicePlatform.ANDROID else None
        )
        self._idb_client = (
            IdbClientWrapper(udid=device_id) if platform == DevicePlatform.IOS else None
        )

    def _run_servers(self, device_id: str, platform: DevicePlatform) -> bool:
        if platform == DevicePlatform.ANDROID:
            if not self._ui_adb_client:
                error_msg = (
                    "UIAutomator client is required for Android but not available. "
                    "Please ensure UIAutomator2 is properly installed and configured."
                )
                logger.error(error_msg)
                raise ValueError(error_msg)
            logger.info("✓ UIAutomator client available for Android")
        elif platform == DevicePlatform.IOS:
            if not self._idb_client:
                error_msg = (
                    "IDB client is required for iOS but not available. "
                    "Please ensure idb companion is properly installed and started."
                )
                logger.error(error_msg)
                raise ValueError(error_msg)
            logger.info("✓ IDB client available for iOS")

        return True

    async def _get_device_context(
        self,
        device_id: str,
        platform: DevicePlatform,
    ) -> DeviceContext:
        from platform import system

        host_platform = system()

        # Get real device dimensions from the device
        if platform == DevicePlatform.ANDROID:
            if self._ui_adb_client:
                try:
                    # Use UIAutomator to get actual screen dimensions
                    screen_data = self._ui_adb_client.get_screen_data()
                    device_width = screen_data.width
                    device_height = screen_data.height
                    logger.info(
                        f"Retrieved Android screen dimensions: {device_width}x{device_height}"
                    )
                except Exception as e:
                    logger.warning(f"Failed to get Android screen dimensions: {e}, using defaults")
                    device_width, device_height = 1080, 2340
            else:
                logger.warning("UIAutomator client not available, using default dimensions")
                device_width, device_height = 1080, 2340
        else:  # iOS
            if self._idb_client:
                try:
                    # Use IDB to take a screenshot and get dimensions
                    screenshot_data = await self._idb_client.screenshot()  # type: ignore[call-arg]
                    if screenshot_data:
                        img = Image.open(BytesIO(screenshot_data))
                        device_width = img.width
                        device_height = img.height
                        logger.info(
                            f"Retrieved iOS screen dimensions: {device_width}x{device_height}"
                        )
                    else:
                        logger.warning("IDB screenshot returned None, using default dimensions")
                        device_width, device_height = 375, 812
                except Exception as e:
                    logger.warning(f"Failed to get iOS screen dimensions: {e}, using defaults")
                    device_width, device_height = 375, 812
            else:
                logger.warning("IDB client not available, using default dimensions")
                device_width, device_height = 375, 812

        return DeviceContext(
            host_platform="WINDOWS" if host_platform == "Windows" else "LINUX",
            mobile_platform=platform,
            device_id=device_id,
            device_width=device_width,
            device_height=device_height,
        )

    def _get_task_status_change_callback(
        self,
        task_info: PlatformTaskInfo,
        platform_service: PlatformService | None = None,
    ) -> Callable[[TaskRunStatus, str | None, Any | None], Coroutine]:
        service = platform_service or self._platform_service

        async def change_status(
            status: TaskRunStatus,
            message: str | None = None,
            output: Any | None = None,
        ):
            if not service:
                raise PlatformServiceUninitializedError()
            try:
                await service.update_task_run_status(
                    task_run_id=task_info.task_run.id,
                    status=status,
                    message=message,
                    output=output,
                )
            except Exception as e:
                logger.error(f"Failed to update task run status: {e}")

        return change_status

    def _get_plan_changes_callback(
        self,
        task_info: PlatformTaskInfo,
        platform_service: PlatformService | None = None,
    ) -> Callable[[list[Subgoal], IsReplan], Coroutine]:
        service = platform_service or self._platform_service
        current_plan: TaskRunPlanResponse | None = None

        async def update_plan(plan: list[Subgoal], is_replan: IsReplan):
            nonlocal current_plan

            if not service:
                raise PlatformServiceUninitializedError()
            try:
                if is_replan and current_plan:
                    # End previous plan
                    await service.upsert_task_run_plan(
                        task_run_id=task_info.task_run.id,
                        started_at=current_plan.started_at,
                        plan=plan,
                        ended_at=datetime.now(UTC),
                        plan_id=current_plan.id,
                    )
                    current_plan = None

                current_plan = await service.upsert_task_run_plan(
                    task_run_id=task_info.task_run.id,
                    started_at=current_plan.started_at if current_plan else datetime.now(UTC),
                    plan=plan,
                    ended_at=current_plan.ended_at if current_plan else None,
                    plan_id=current_plan.id if current_plan else None,
                )
            except Exception as e:
                logger.error(f"Failed to update plan: {e}")

        return update_plan

    def _get_new_agent_thought_callback(
        self,
        task_info: PlatformTaskInfo,
        platform_service: PlatformService | None = None,
    ) -> Callable[[AgentNode, str], Coroutine]:
        service = platform_service or self._platform_service

        async def add_agent_thought(agent: AgentNode, thought: str):
            if not service:
                raise PlatformServiceUninitializedError()
            try:
                await service.add_agent_thought(
                    task_run_id=task_info.task_run.id,
                    agent=agent,
                    thought=thought,
                )
            except Exception as e:
                logger.error(f"Failed to add agent thought: {e}")

        return add_agent_thought


def _validate_and_prepare_file(file_path: Path):
    path_obj = Path(file_path)
    if path_obj.exists() and path_obj.is_dir():
        raise AgentTaskRequestError(f"Error: Path '{file_path}' is a directory, not a file.")
    try:
        path_obj.parent.mkdir(parents=True, exist_ok=True)
        path_obj.touch(exist_ok=True)
    except OSError as e:
        raise AgentTaskRequestError(f"Error creating file '{file_path}': {e}")


def print_ai_response_to_stderr(graph_result: State):
    for msg in reversed(graph_result.messages):
        if isinstance(msg, AIMessage):
            print(msg.content, file=sys.stderr)
            return



import asyncio
import os
from shutil import which
from typing import Annotated

import typer
from adbutils import AdbClient
from langchain.callbacks.base import Callbacks
from rich.console import Console

from minitap.mobile_use.config import initialize_llm_config, settings
from minitap.mobile_use.sdk import Agent
from minitap.mobile_use.sdk.builders import Builders
from minitap.mobile_use.sdk.types.task import AgentProfile
from minitap.mobile_use.utils.cli_helpers import display_device_status
from minitap.mobile_use.utils.logger import get_logger

app = typer.Typer(add_completion=False, pretty_exceptions_enable=False)
logger = get_logger(__name__)


async def run_automation(
    goal: str,
    locked_app_package: str | None = None,
    test_name: str | None = None,
    traces_output_path_str: str = "traces",
    output_description: str | None = None,
    graph_config_callbacks: Callbacks = [],
):
    llm_config = initialize_llm_config()
    agent_profile = AgentProfile(name="default", llm_config=llm_config)
    config = Builders.AgentConfig.with_default_profile(profile=agent_profile)

    if settings.ADB_HOST:
        config.with_adb_server(host=settings.ADB_HOST, port=settings.ADB_PORT)
    if graph_config_callbacks:
        config.with_graph_config_callbacks(graph_config_callbacks)

    agent = Agent(config=config.build())
    await agent.init(
        retry_count=int(os.getenv("MOBILE_USE_HEALTH_RETRIES", 5)),
        retry_wait_seconds=int(os.getenv("MOBILE_USE_HEALTH_DELAY", 2)),
    )

    task = agent.new_task(goal)
    if locked_app_package:
        task.with_locked_app_package(locked_app_package)
    if test_name:
        task.with_name(test_name).with_trace_recording(path=traces_output_path_str)
    if output_description:
        task.with_output_description(output_description)

    agent_thoughts_path = os.getenv("EVENTS_OUTPUT_PATH", None)
    llm_result_path = os.getenv("RESULTS_OUTPUT_PATH", None)
    if agent_thoughts_path:
        task.with_thoughts_output_saving(path=agent_thoughts_path)
    if llm_result_path:
        task.with_llm_output_saving(path=llm_result_path)

    await agent.run_task(request=task.build())

    await agent.clean()


@app.command()
def main(
    goal: Annotated[str, typer.Argument(help="The main goal for the agent to achieve.")],
    test_name: Annotated[
        str | None,
        typer.Option(
            "--test-name",
            "-n",
            help="A name for the test recording. If provided, a trace will be saved.",
        ),
    ] = None,
    traces_path: Annotated[
        str,
        typer.Option(
            "--traces-path",
            "-p",
            help="The path to save the traces.",
        ),
    ] = "traces",
    output_description: Annotated[
        str | None,
        typer.Option(
            "--output-description",
            "-o",
            help=(
                """
                A dict output description for the agent.
                Ex: a JSON schema with 2 keys: type, price
                """
            ),
        ),
    ] = None,
):
    """
    Run the Mobile-use agent to automate tasks on a mobile device.
    """
    console = Console()

    adb_client = None
    try:
        if which("adb"):
            adb_client = AdbClient(
                host=settings.ADB_HOST or "localhost",
                port=settings.ADB_PORT or 5037,
            )
    except Exception:
        pass  # ADB not available, will only support iOS devices

    display_device_status(console, adb_client=adb_client)
    asyncio.run(
        run_automation(
            goal=goal,
            test_name=test_name,
            traces_output_path_str=traces_path,
            output_description=output_description,
        )
    )


def cli():
    app()


if __name__ == "__main__":
    cli()



"""iOS-specific device controller implementation using IDB."""

import asyncio
import base64
import re
from io import BytesIO

from idb.common.types import HIDButtonType
from PIL import Image

from minitap.mobile_use.clients.idb_client import IdbClientWrapper
from minitap.mobile_use.controllers.device_controller import (
    MobileDeviceController,
    ScreenDataResponse,
)
from minitap.mobile_use.controllers.types import Bounds, CoordinatesSelectorRequest, TapOutput
from minitap.mobile_use.utils.logger import get_logger

logger = get_logger(__name__)


class iOSDeviceController(MobileDeviceController):
    """iOS device controller using IDB."""

    def __init__(
        self,
        idb_client: IdbClientWrapper,
        device_width: int,
        device_height: int,
    ):
        self.idb_client = idb_client
        self.device_width = device_width
        self.device_height = device_height

    async def tap(
        self,
        coords: CoordinatesSelectorRequest,
        long_press: bool = False,
        long_press_duration: int = 1000,
    ) -> TapOutput:
        """Tap at specific coordinates using IDB."""
        try:
            duration = long_press_duration / 1000.0 if long_press else None
            await self.idb_client.tap(x=coords.x, y=coords.y, duration=duration)  # type: ignore[call-arg]
            return TapOutput(error=None)
        except Exception as e:
            return TapOutput(error=f"IDB tap failed: {str(e)}")

    async def swipe(
        self,
        start: CoordinatesSelectorRequest,
        end: CoordinatesSelectorRequest,
        duration: int = 400,
    ) -> str | None:
        """Swipe from start to end coordinates using IDB."""
        try:
            # IDB delta is the number of steps, approximating from duration
            ms_duration_to_percentage = duration / 1000.0
            await self.idb_client.swipe(  # type: ignore[call-arg]
                x_start=start.x,
                y_start=start.y,
                x_end=end.x,
                y_end=end.y,
                duration=ms_duration_to_percentage,
            )
            return None
        except Exception as e:
            return f"IDB swipe failed: {str(e)}"

    async def get_screen_data(self) -> ScreenDataResponse:
        """Get screen data using IDB (screenshot and hierarchy in parallel)."""
        try:
            # Run screenshot and hierarchy fetch in parallel
            screenshot_bytes, accessibility_info = await asyncio.gather(
                self.idb_client.screenshot(),  # type: ignore[call-arg]
                self.idb_client.describe_all(),
            )

            if screenshot_bytes is None:
                raise RuntimeError("Screenshot returned None")

            base64_screenshot = base64.b64encode(screenshot_bytes).decode("utf-8")
            elements = (
                self._process_flat_ios_hierarchy(accessibility_info) if accessibility_info else []
            )

            return ScreenDataResponse(
                base64=base64_screenshot,
                elements=elements,
                width=self.device_width,
                height=self.device_height,
                platform="ios",
            )
        except Exception as e:
            logger.error(f"Failed to get screen data: {e}")
            raise

    async def screenshot(self) -> str:
        """Take a screenshot using IDB and return base64 encoded string."""
        try:
            screenshot_bytes = await self.idb_client.screenshot()  # type: ignore[call-arg]
            if screenshot_bytes is None:
                raise RuntimeError("Screenshot returned None")
            return base64.b64encode(screenshot_bytes).decode("utf-8")
        except Exception as e:
            logger.error(f"Failed to take screenshot: {e}")
            raise

    async def input_text(self, text: str) -> bool:
        """Input text using IDB."""
        try:
            return await self.idb_client.text(text)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to input text: {e}")
            return False

    async def launch_app(self, package_or_bundle_id: str) -> bool:
        """Launch an iOS app using IDB."""
        try:
            return await self.idb_client.launch(bundle_id=package_or_bundle_id)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to launch app {package_or_bundle_id}: {e}")
            return False

    async def terminate_app(self, package_or_bundle_id: str | None) -> bool:
        """Terminate an iOS app using IDB."""
        if package_or_bundle_id is None:
            logger.warning("Cannot terminate app: bundle_id is None")
            return False
        try:
            return await self.idb_client.terminate(bundle_id=package_or_bundle_id)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to terminate app {package_or_bundle_id}: {e}")
            return False

    async def open_url(self, url: str) -> bool:
        """Open a URL using IDB."""
        try:
            return await self.idb_client.open_url(url)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to open URL {url}: {e}")
            return False

    async def press_back(self) -> bool:
        """iOS doesn't have a back button - swipe from left edge."""
        try:
            # Simulate back gesture by swiping from left edge
            start = CoordinatesSelectorRequest(x=10, y=self.device_height // 4)
            end = CoordinatesSelectorRequest(x=300, y=self.device_height // 4)
            result = await self.swipe(start, end, duration=300)
            return result is None
        except Exception as e:
            logger.error(f"Failed to press back: {e}")
            return False

    async def press_home(self) -> bool:
        """Press the home button using IDB."""
        try:
            return await self.idb_client.button(button_type=HIDButtonType.HOME)  # type: ignore[call-arg]
        except Exception as e:
            logger.error(f"Failed to press home: {e}")
            return False

    async def get_ui_hierarchy(self) -> list[dict]:
        """Get UI hierarchy using IDB accessibility info."""
        try:
            accessibility_info = await asyncio.wait_for(
                self.idb_client.describe_all(), timeout=40.0
            )
            if accessibility_info is None:
                logger.warning("Accessibility info returned None")
                return []

            hierarchy = self._process_flat_ios_hierarchy(accessibility_info)
            return hierarchy
        except TimeoutError:
            logger.error("Timeout waiting for UI hierarchy (40s)")
            return []
        except Exception as e:
            logger.error(f"Failed to get UI hierarchy: {e}")
            return []

    def _process_flat_ios_hierarchy(self, accessibility_data: list[dict]) -> list[dict]:
        """
        Process flat iOS accessibility info into our standard format.

        IDB with nested=False returns a flat list of all elements.
        """
        elements = []

        for node in accessibility_data:
            if not isinstance(node, dict):
                continue

            # Extract element info
            element = {
                "type": node.get("type", ""),
                "value": node.get("AXValue", ""),
                "label": node.get("AXLabel", node.get("label", "")),
                "frame": node.get("frame", {}),
                "enabled": node.get("enabled", False),
                "visible": True,  # Elements in the list are generally visible
            }

            # Add bounds if frame is available
            if "frame" in node and isinstance(node["frame"], dict):
                frame = node["frame"]
                if all(k in frame for k in ["x", "y", "width", "height"]):
                    element["bounds"] = (
                        f"[{int(frame['x'])},{int(frame['y'])}]"
                        f"[{int(frame['x'] + frame['width'])},{int(frame['y'] + frame['height'])}]"
                    )

            elements.append(element)

        return elements

    def find_element(
        self,
        ui_hierarchy: list[dict],
        resource_id: str | None = None,
        text: str | None = None,
        index: int = 0,
    ) -> tuple[dict | None, Bounds | None, str | None]:
        """Find a UI element in the iOS hierarchy."""
        if not resource_id and not text:
            return None, None, "No resource_id or text provided"

        matches = []
        for element in ui_hierarchy:
            # iOS doesn't have resource-id, so we match on type if provided as resource_id
            if resource_id and element.get("type") == resource_id:
                matches.append(element)
            # Match on value or label for text
            elif text and (element.get("value") == text or element.get("label") == text):
                matches.append(element)

        if not matches:
            criteria = f"type='{resource_id}'" if resource_id else f"text='{text}'"
            return None, None, f"No element found with {criteria}"

        if index >= len(matches):
            criteria = f"type='{resource_id}'" if resource_id else f"text='{text}'"
            return (
                None,
                None,
                f"Index {index} out of range for {criteria} (found {len(matches)} matches)",
            )

        element = matches[index]
        bounds = self._extract_bounds(element)

        return element, bounds, None

    def _extract_bounds(self, element: dict) -> Bounds | None:
        """Extract bounds from an iOS UI element."""
        bounds_str = element.get("bounds")
        if not bounds_str or not isinstance(bounds_str, str):
            return None

        try:
            # Parse bounds string like "[x1,y1][x2,y2]"
            match = re.match(r"\[(\d+),(\d+)\]\[(\d+),(\d+)\]", bounds_str)
            if match:
                return Bounds(
                    x1=int(match.group(1)),
                    y1=int(match.group(2)),
                    x2=int(match.group(3)),
                    y2=int(match.group(4)),
                )
        except (ValueError, IndexError):
            return None

        return None

    async def erase_text(self, nb_chars: int | None = None) -> bool:
        """Erase text by sending delete key presses."""
        try:
            if nb_chars is None:
                nb_chars = 50  # Default to erasing 50 characters
            # iOS delete key code is 42 (HID keyboard delete)
            for _ in range(nb_chars):
                await self.idb_client.key(42)  # type: ignore[call-arg]
            return True
        except Exception as e:
            logger.error(f"Failed to erase text: {e}")
            return False

    async def cleanup(self) -> None:
        """Cleanup iOS controller resources."""
        logger.debug("iOS controller cleanup")
        await self.idb_client.cleanup()

    def get_compressed_b64_screenshot(self, image_base64: str, quality: int = 50) -> str:
        if image_base64.startswith("data:image"):
            image_base64 = image_base64.split(",")[1]

        image_data = base64.b64decode(image_base64)
        image = Image.open(BytesIO(image_data))

        # Convert RGBA to RGB if image has alpha channel (PNG transparency)
        if image.mode in ("RGBA", "LA", "P"):
            rgb_image = Image.new("RGB", image.size, (255, 255, 255))
            rgb_image.paste(image, mask=image.split()[-1] if image.mode == "RGBA" else None)
            image = rgb_image

        compressed_io = BytesIO()
        image.save(compressed_io, format="JPEG", quality=quality, optimize=True)

        compressed_base64 = base64.b64encode(compressed_io.getvalue()).decode("utf-8")
        return compressed_base64



import json
import re
from datetime import date
from shutil import which

from adbutils import AdbDevice

from minitap.mobile_use.context import DevicePlatform, MobileUseContext
from minitap.mobile_use.utils.logger import MobileUseLogger, get_logger
from minitap.mobile_use.utils.shell_utils import run_shell_command_on_host

logger = get_logger(__name__)


def get_adb_device(ctx: MobileUseContext) -> AdbDevice:
    if ctx.device.mobile_platform != DevicePlatform.ANDROID:
        raise ValueError("Device is not an Android device")
    adb = ctx.get_adb_client()
    device = adb.device(serial=ctx.device.device_id)
    if not device:
        raise ConnectionError(f"Device {ctx.device.device_id} not found.")
    return device


def get_first_device(
    logger: MobileUseLogger | None = None,
) -> tuple[str | None, DevicePlatform | None]:
    """Gets the first available device."""
    if which("adb"):
        try:
            android_output = run_shell_command_on_host("adb devices")
            lines = android_output.strip().split("\n")
            for line in lines:
                if "device" in line and not line.startswith("List of devices"):
                    return line.split()[0], DevicePlatform.ANDROID
        except RuntimeError as e:
            if logger:
                logger.error(f"ADB command failed: {e}")

    if which("xcrun"):
        try:
            ios_output = run_shell_command_on_host("xcrun simctl list devices booted -j")
            data = json.loads(ios_output)
            for runtime, devices in data.get("devices", {}).items():
                if "iOS" not in runtime:
                    continue
                for device in devices:
                    if device.get("state") == "Booted":
                        return device["udid"], DevicePlatform.IOS
        except RuntimeError as e:
            if logger:
                logger.error(f"xcrun command failed: {e}")

    return None, None


def get_device_date(ctx: MobileUseContext) -> str:
    if ctx.device.mobile_platform == DevicePlatform.IOS:
        return date.today().strftime("%a %b %d %H:%M:%S %Z %Y")
    device = get_adb_device(ctx)
    return str(device.shell("date"))


def list_packages(ctx: MobileUseContext) -> str:
    if ctx.device.mobile_platform == DevicePlatform.IOS:
        cmd = ["xcrun", "simctl", "listapps", "booted", "|", "grep", "CFBundleIdentifier"]
        return run_shell_command_on_host(" ".join(cmd))
    else:
        device = get_adb_device(ctx)
        # Get full package list with paths
        cmd = ["pm", "list", "packages", "-f"]
        raw_output = str(device.shell(" ".join(cmd)))

        # Extract only package names (remove paths and "package:" prefix)
        # Format: "package:/path/to/app.apk=com.example.app" -> "com.example.app"
        lines = raw_output.strip().split("\n")
        packages = []
        for line in lines:
            if "=" in line:
                package_name = line.split("=")[-1].strip()
                packages.append(package_name)

        return "\n".join(sorted(packages))


def get_current_foreground_package(ctx: MobileUseContext) -> str | None:
    """
    Get the package name of the currently focused/foreground app.

    Returns only the clean package/bundle name (e.g., 'com.whatsapp'),
    without any metadata or window information.

    Returns:
        The package/bundle name, or None if unable to determine
    """
    try:
        if ctx.device.mobile_platform == DevicePlatform.IOS:
            output = run_shell_command_on_host(
                "xcrun simctl spawn booted launchctl print "
                "system/com.apple.SpringBoard.services | grep bundleIdentifier"
            )
            match = re.search(r'"bundleIdentifier"\s*=\s*"([^"]+)"', output)
            if match:
                bundle_id = match.group(1)
                if "." in bundle_id:
                    return bundle_id
            return None

        device = get_adb_device(ctx)
        output = str(device.shell("dumpsys window | grep mCurrentFocus"))

        if "mCurrentFocus=" not in output:
            return None

        segment = output.split("mCurrentFocus=")[-1]

        if "/" in segment:
            tokens = segment.split()
            for token in tokens:
                if "." in token and not token.startswith("Window"):
                    package = token.split("/")[0]
                    package = package.rstrip("}")
                    if package and "." in package:
                        return package

        return None

    except Exception as e:
        logger.debug(f"Failed to get current foreground package: {e}")
        return None



import json
from pathlib import Path

from PIL import Image


def quantize_and_save_gif(
    images: list[Image.Image],
    output_path: Path,
    colors: int = 128,
    duration: int = 100,
) -> None:
    """
    Quantize images and save as an optimized GIF.

    Args:
        images: List of PIL Image objects to convert to GIF
        output_path: Path where the GIF will be saved
        colors: Number of colors to use in quantization (lower = smaller file)
        duration: Duration of each frame in milliseconds

    Raises:
        ValueError: If images list is empty
    """
    if not images:
        raise ValueError("images must not be empty")

    quantized_images = []
    for img in images:
        if img.mode != "RGB":
            img = img.convert("RGB")
        quantized = img.quantize(colors=colors, method=2)
        quantized_images.append(quantized)

    quantized_images[0].save(
        output_path,
        save_all=True,
        append_images=quantized_images[1:],
        loop=0,
        optimize=True,
        duration=duration,
    )


def create_gif_from_trace_folder(trace_folder_path: Path):
    images = []
    image_files = []

    for file in trace_folder_path.iterdir():
        if file.suffix == ".jpeg":
            image_files.append(file)

    image_files.sort(key=lambda f: int(f.stem))

    print("Found " + str(len(image_files)) + " images to compile")

    for file in image_files:
        with open(file, "rb") as f:
            image = Image.open(f).convert("RGB")
            images.append(image)

    if len(images) == 0:
        return

    gif_path = trace_folder_path / "trace.gif"
    quantize_and_save_gif(images, gif_path)
    print("GIF created at " + str(gif_path))


def remove_images_from_trace_folder(trace_folder_path: Path):
    for file in trace_folder_path.iterdir():
        if file.suffix == ".jpeg":
            file.unlink()


def create_steps_json_from_trace_folder(trace_folder_path: Path):
    steps = []
    for file in trace_folder_path.iterdir():
        if file.suffix == ".json":
            with open(file, encoding="utf-8", errors="ignore") as f:
                json_content = f.read()
                steps.append({"timestamp": int(file.stem), "data": json_content})

    steps.sort(key=lambda f: f["timestamp"])

    print("Found " + str(len(steps)) + " steps to compile")

    with open(trace_folder_path / "steps.json", "w", encoding="utf-8", errors="ignore") as f:
        f.write(json.dumps(steps))


def remove_steps_json_from_trace_folder(trace_folder_path: Path):
    for file in trace_folder_path.iterdir():
        if file.suffix == ".json" and file.name != "steps.json":
            file.unlink()



import asyncio
import json
import socket
import subprocess
from functools import wraps
from pathlib import Path
from typing import Any

from idb.common.types import HIDButtonType, InstalledAppInfo, InstalledArtifact, TCPAddress
from idb.grpc.client import Client

from minitap.mobile_use.utils.logger import get_logger

logger = get_logger(__name__)


def _find_available_port(start_port: int = 10882, max_attempts: int = 100) -> int:
    for port in range(start_port, start_port + max_attempts):
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            try:
                s.bind(("localhost", port))
                return port
            except OSError:
                continue
    raise RuntimeError(
        f"Could not find available port in range {start_port}-{start_port + max_attempts}"
    )


def with_idb_client(func):
    """Decorator to handle idb client lifecycle.

    Note: Function must have None or bool in return type.
    """

    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        method_name = func.__name__
        try:
            logger.debug(f"Building IDB client connection for {method_name}...")
            async with Client.build(address=self.address, logger=logger.logger) as client:
                logger.debug(f"Client built, calling {method_name}...")
                result = await func(self, client, *args, **kwargs)
                logger.debug(f"{method_name} completed successfully")
                return result
        except Exception as e:
            logger.error(f"Failed to {method_name}: {e}")
            import traceback

            logger.debug(f"Traceback: {traceback.format_exc()}")

            return_type = func.__annotations__.get("return")
            if return_type is bool:
                return False
            return None

    return wrapper


class IdbClientWrapper:
    """Wrapper around fb-idb client for iOS device automation with lifecycle management.

    This wrapper can either manage the idb_companion process lifecycle locally or connect
    to an external companion server.

    Lifecycle Management:
    - If host is None (default): Manages companion locally on localhost
      - Call init_companion() to start the idb_companion process
      - Call cleanup() to stop the companion process
      - Or use as async context manager for automatic lifecycle
    - If host is provided: Connects to external companion server
      - init_companion() and cleanup() become no-ops
      - You manage the external companion separately

    Example:
        # Managed companion (recommended for local development)
        async with IdbClientWrapper(udid="device-id") as wrapper:
            await wrapper.tap(100, 200)

        # External companion (for production/remote)
        wrapper = IdbClientWrapper(udid="device-id", host="remote-host", port=10882)
        await wrapper.tap(100, 200)  # No companion lifecycle management needed
    """

    def __init__(self, udid: str, host: str | None = None, port: int | None = None):
        self.udid = udid
        self._manage_companion = host is None

        if host is None:
            actual_port = port if port is not None else _find_available_port()
            self.address = TCPAddress(host="localhost", port=actual_port)
            logger.debug(f"Will manage companion for {udid} on port {actual_port}")
        else:
            actual_port = port if port is not None else 10882
            self.address = TCPAddress(host=host, port=actual_port)

        self.companion_process: subprocess.Popen | None = None

    async def init_companion(self, idb_companion_path: str = "idb_companion") -> bool:
        """
        Start the idb_companion process for this device.
        Only starts if managing companion locally (host was None in __init__).

        Args:
            idb_companion_path: Path to idb_companion binary (default: "idb_companion" from PATH)

        Returns:
            True if companion started successfully, False otherwise
        """
        if not self._manage_companion:
            logger.info(f"Using external idb_companion at {self.address.host}:{self.address.port}")
            return True

        if self.companion_process is not None:
            logger.warning(f"idb_companion already running for {self.udid}")
            return True

        try:
            cmd = [idb_companion_path, "--udid", self.udid, "--grpc-port", str(self.address.port)]

            logger.info(f"Starting idb_companion: {' '.join(cmd)}")
            self.companion_process = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True
            )

            # Wait longer for gRPC server to be fully ready
            logger.debug("Waiting for idb_companion gRPC server to be ready...")
            await asyncio.sleep(5)

            if self.companion_process.poll() is not None:
                stdout, stderr = self.companion_process.communicate()
                logger.error(f"idb_companion failed to start: {stderr}")
                self.companion_process = None
                return False

            logger.info(
                f"idb_companion started successfully for {self.udid} on port {self.address.port}"
            )
            return True

        except FileNotFoundError:
            logger.error(
                "idb_companion not found. Please install fb-idb to use iOS devices.\n"
                "Installation guide: https://fbidb.io/docs/installation/\n"
                "On macOS with Homebrew: brew install idb-companion"
            )
            self.companion_process = None
            return False
        except Exception as e:
            logger.error(f"Failed to start idb_companion: {e}")
            self.companion_process = None
            return False

    async def cleanup(self) -> None:
        if not self._manage_companion:
            logger.debug(f"Not managing companion for {self.udid}, skipping cleanup")
            return

        if self.companion_process is None:
            return

        try:
            logger.info(f"Stopping idb_companion for {self.udid}")

            self.companion_process.terminate()

            try:
                await asyncio.wait_for(asyncio.to_thread(self.companion_process.wait), timeout=5.0)
                logger.info(f"idb_companion stopped gracefully for {self.udid}")
            except TimeoutError:
                logger.warning(f"Force killing idb_companion for {self.udid}")
                self.companion_process.kill()
                await asyncio.to_thread(self.companion_process.wait)

        except Exception as e:
            logger.error(f"Error stopping idb_companion: {e}")
        finally:
            self.companion_process = None

    def __del__(self):
        if self.companion_process is not None:
            try:
                self.companion_process.terminate()
                self.companion_process.wait(timeout=2)
            except Exception:
                try:
                    self.companion_process.kill()
                except Exception:
                    pass

    async def __aenter__(self):
        if not await self.init_companion():
            raise RuntimeError(f"Failed to initialize idb_companion for device {self.udid}")
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.cleanup()
        return False

    @with_idb_client
    async def tap(self, client: Client, x: int, y: int, duration: float | None = None) -> bool:
        await client.tap(x=x, y=y, duration=duration)
        return True

    @with_idb_client
    async def swipe(
        self,
        client: Client,
        x_start: int,
        y_start: int,
        x_end: int,
        y_end: int,
        duration: float | None = None,
    ) -> bool:
        await client.swipe(p_start=(x_start, y_start), p_end=(x_end, y_end), duration=duration)
        return True

    @with_idb_client
    async def screenshot(self, client: Client, output_path: str | None = None) -> bytes | None:
        """
        Take a screenshot and return raw image data.

        Returns:
            Raw image data (PNG bytes not base64 encoded)
        """
        screenshot_data = await client.screenshot()
        if output_path:
            with open(output_path, "wb") as f:
                f.write(screenshot_data)
        return screenshot_data

    @with_idb_client
    async def launch(
        self,
        client: Client,
        bundle_id: str,
        args: list[str] | None = None,
        env: dict[str, str] | None = None,
    ) -> bool:
        await client.launch(
            bundle_id=bundle_id, args=args or [], env=env or {}, foreground_if_running=True
        )
        return True

    @with_idb_client
    async def terminate(self, client: Client, bundle_id: str) -> bool:
        await client.terminate(bundle_id)
        return True

    @with_idb_client
    async def install(self, client: Client, app_path: str) -> list[InstalledArtifact] | None:
        bundle_path = Path(app_path)
        artifacts = []
        with open(bundle_path, "rb") as f:
            async for artifact in client.install(bundle=f):
                artifacts.append(artifact)
        return artifacts

    @with_idb_client
    async def uninstall(self, client: Client, bundle_id: str) -> bool:
        await client.uninstall(bundle_id)
        return True

    @with_idb_client
    async def list_apps(self, client: Client) -> list[InstalledAppInfo] | None:
        apps = await client.list_apps()
        return apps

    @with_idb_client
    async def text(self, client: Client, text: str) -> bool:
        await client.text(text)
        return True

    @with_idb_client
    async def key(self, client: Client, key_code: int) -> bool:
        await client.key(key_code)
        return True

    @with_idb_client
    async def button(self, client: Client, button_type: HIDButtonType) -> bool:
        await client.button(button_type=button_type)
        return True

    @with_idb_client
    async def clear_keychain(self, client: Client) -> bool:
        await client.clear_keychain()
        return True

    @with_idb_client
    async def open_url(self, client: Client, url: str) -> bool:
        await client.open_url(url)
        return True

    async def describe_all(self) -> list[dict[str, Any]] | None:
        try:
            cmd = ["idb", "ui", "describe-all", "--udid", self.udid, "--json"]
            process = await asyncio.create_subprocess_exec(
                *cmd,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            stdout, stderr = await process.communicate()

            if process.returncode != 0:
                logger.error(f"idb describe-all failed: {stderr.decode()}")
                return None

            parsed = json.loads(stdout.decode())
            return parsed if isinstance(parsed, list) else [parsed]
        except Exception as e:
            logger.error(f"Failed to describe_all: {e}")
            return None

    @with_idb_client
    async def describe_point(self, client: Client, x: int, y: int) -> dict[str, Any] | None:
        accessibility_info = await client.accessibility_info(point=(x, y), nested=True)
        return json.loads(accessibility_info.json)
